@@ -48,7 +48,9 @@ from nwg_panel.modules.scratchpad import Scratchpad
 from nwg_panel.modules.dwl_tags import DwlTags
 from nwg_panel.modules.swaync import SwayNC
 from nwg_panel.modules.sway_mode import SwayMode
+from nwg_panel.modules.hyprland_submap import HyprlandSubmap
 from nwg_panel.modules.keyboard_layout import KeyboardLayout
+from nwg_panel.modules.random_wallpaper import RandomWallpaper
 
 try:
     from nwg_panel.modules.openweather import OpenWeather
@@ -87,8 +89,6 @@ his = os.getenv('HYPRLAND_INSTANCE_SIGNATURE')
 if his:
     from nwg_panel.modules.hyprland_taskbar import HyprlandTaskbar
     from nwg_panel.modules.hyprland_workspaces import HyprlandWorkspaces
-last_client_addr = ""
-last_client_title = ""
 
 common_settings = {}
 restart_cmd = ""
@@ -154,80 +154,62 @@ def restart():
     subprocess.Popen(restart_cmd, shell=True)
 
 
-# read from Hyprland socket2 on async thread
 def hypr_watcher():
     import socket
 
-    client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-    client.connect("/tmp/hypr/{}/.socket2.sock".format(his))
+    # /tmp/hypr moved to $XDG_RUNTIME_DIR/hypr in #5788
+    xdg_runtime_dir = os.getenv("XDG_RUNTIME_DIR")
+    hypr_dir = f"{xdg_runtime_dir}/hypr" if xdg_runtime_dir and os.path.isdir(
+        f"{xdg_runtime_dir}/hypr") else "/tmp/hypr"
 
-    global last_client_addr, last_client_title
-    client_addr, client_title = None, None
+    client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+    client.connect(f"{hypr_dir}/{his}/.socket2.sock")
+    just_refreshed = False
 
     while True:
         datagram = client.recv(2048)
-        e_full_string = datagram.decode('utf-8').strip()
-        # eprint("Event: {}".format(e_full_string))
+        e_full_string = datagram.decode('utf-8', errors='replace').strip()
+        lines = e_full_string.splitlines()
 
-        # remember client address & title (string) for further event filtering
-        if e_full_string.startswith("activewindow"):
-            lines = e_full_string.splitlines()
-            for line in lines:
-                if line.startswith("activewindowv2"):
-                    client_addr = e_full_string.split(">>")[1].strip()
-                elif line.startswith("activewindow>>"):
-                    client_title = line.split(">>")[1].strip()
+        event_names = []
+        for line in lines:
+            event_names.append(line.split(">>")[0])
+        # print(f"events: {event_names}")
 
-        event_name = e_full_string.split(">>")[0]
+        for event_name in event_names:
+            if event_name in ["activespecial",
+                              "activewindow",
+                              "activewindowv2",
+                              "changefloatingmode",
+                              "closewindow",
+                              "createworkspace",
+                              "destroyworkspace",
+                              "focusedmon",
+                              "monitoradded",
+                              "movewindow",
+                              "openwindow",
+                              "windowtitle",
+                              "workspace"]:
 
-        if event_name in ["monitoradded", "openwindow", "movewindow"]:
-            monitors, workspaces, clients, activewindow = h_modules_get_all()
-            for item in common.h_taskbars_list:
-                GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow)
-            last_client_title = client_title
-            last_client_addr = client_addr
-            continue
+                if "activewindow" in event_name and just_refreshed:
+                    just_refreshed = False
+                    break
 
-        if event_name == "focusedmon":
-            monitors, workspaces, clients, activewindow = h_modules_get_all()
-            for item in common.h_workspaces_list:
-                GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow)
-            last_client_title = client_title
-            last_client_addr = client_addr
-            continue
+                # print(f">>> refreshing on {event_name}")
+                monitors, workspaces, clients, activewindow, activeworkspace = h_modules_get_all()
+                for item in common.h_taskbars_list:
+                    GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow)
 
-        if event_name == "activewindow" and client_title != last_client_title:
-            monitors, workspaces, clients, activewindow = h_modules_get_all()
-            for item in common.h_taskbars_list:
-                GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow)
+                for item in common.h_workspaces_list:
+                    GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow, activeworkspace)
 
-            for item in common.h_workspaces_list:
-                GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow)
+                if event_name in ["createworkspace", "destroyworkspace", "focusedmon", "workspace"]:
+                    just_refreshed = True
+                break
 
-            last_client_title = client_title
-            continue
-
-        if event_name == "activewindowv2" and client_addr != last_client_addr:
-            monitors, workspaces, clients, activewindow = h_modules_get_all()
-            for item in common.h_taskbars_list:
-                GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow)
-
-            for item in common.h_workspaces_list:
-                GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow)
-
-            last_client_addr = client_addr
-            continue
-
-        if event_name in ["changefloatingmode", "closewindow"]:
-            monitors, workspaces, clients, activewindow = h_modules_get_all()
-            for item in common.h_taskbars_list:
-                GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow)
-
-            for item in common.h_workspaces_list:
-                GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow)
-
-            last_client_addr = ""
-            last_client_title = ""
+            elif event_name == "submap":
+                for item in common.h_submaps_list:
+                    GLib.timeout_add(0, item.refresh)
 
 
 def on_i3ipc_event(i3conn, event):
@@ -262,14 +244,14 @@ def instantiate_content(panel, container, content_list, icons_path=""):
     check_key(panel, "position", "top")
     check_key(panel, "items-padding", 0)
 
-    # list initial data for Hyprland modules
-    if his:
-        if "hyprland-workspaces" in content_list or "hyprland-taskbar" in content_list:
-            monitors, workspaces, clients, activewindow = h_modules_get_all()
-        else:
-            monitors, workspaces, clients, activewindow = {}, {}, {}, {}
-
     for item in content_list:
+        # list initial data for Hyprland modules
+        if his:
+            if "hyprland-workspaces" in content_list or "hyprland-taskbar" in content_list:
+                monitors, workspaces, clients, activewindow, activeworkspace = h_modules_get_all()
+            else:
+                monitors, workspaces, clients, activewindow, activeworkspace = {}, {}, {}, {}, {}
+
         if item == "sway-taskbar":
             if "sway-taskbar" in panel:
                 if sway:
@@ -319,11 +301,11 @@ def instantiate_content(panel, container, content_list, icons_path=""):
 
         if item == "hyprland-taskbar":
             if "hyprland-taskbar" in panel:
-                if panel["layer"] in ["bottom", "background"]:
-                    eprint(
-                        "Panel '{}': On Hyprland, panels must be placed on 'top' or 'overlay' layer, but '{}' found in "
-                        "settings. Changing to 'top'.".format(panel["name"], panel["layer"]))
-                    panel["layer"] = "top"  # or context menu will remain invisible
+                # if panel["layer"] in ["bottom", "background"]:
+                #     eprint(
+                #         "Panel '{}': On Hyprland, panels must be placed on 'top' or 'overlay' layer, but '{}' found in "
+                #         "settings. Changing to 'top'.".format(panel["name"], panel["layer"]))
+                #     panel["layer"] = "top"  # or context menu will remain invisible
                 if his:
                     check_key(panel["hyprland-taskbar"], "all-outputs", False)
                     if panel["hyprland-taskbar"]["all-outputs"] or "output" not in panel:
@@ -343,7 +325,7 @@ def instantiate_content(panel, container, content_list, icons_path=""):
             if his:
                 if "hyprland-workspaces" in panel:
                     workspaces = HyprlandWorkspaces(panel["hyprland-workspaces"], monitors, workspaces, clients,
-                                                    activewindow, icons_path=icons_path)
+                                                    activewindow, activeworkspace, icons_path=icons_path)
                     container.pack_start(workspaces, False, False, panel["items-padding"])
                     common.h_workspaces_list.append(workspaces)
                 else:
@@ -351,6 +333,17 @@ def instantiate_content(panel, container, content_list, icons_path=""):
             else:
                 eprint("'hyprland-workspaces' ignored")
 
+        if item == "hyprland-submap":
+            if his:
+                if item in panel:
+                    h_submap = HyprlandSubmap(panel[item], icons_path=icons_path)
+                else:
+                    h_submap = HyprlandSubmap({}, icons_path=icons_path)
+                container.pack_start(h_submap, False, False, panel["items-padding"])
+                common.h_submaps_list.append(h_submap)
+            else:
+                eprint("HyprlandSubmap module only works on Hyprland, ignoring")
+
         if item == "keyboard-layout":
             if his or sway:
                 if "keyboard-layout" not in panel:
@@ -385,7 +378,7 @@ def instantiate_content(panel, container, content_list, icons_path=""):
 
         if item == "playerctl":
             if item in panel:
-                playerctl = Playerctl(panel[item], icons_path)
+                playerctl = Playerctl(panel[item], voc, icons_path)
                 container.pack_start(playerctl, False, False, panel["items-padding"])
             else:
                 print("'{}' not defined in this panel instance".format(item))
@@ -407,6 +400,12 @@ def instantiate_content(panel, container, content_list, icons_path=""):
             cpu_avg = CpuAvg()
             container.pack_start(cpu_avg, False, False, panel["items-padding"])
 
+        if item == "random-wallpaper":
+            if item not in panel:
+                panel[item] = {}
+            cpu_avg = RandomWallpaper(panel[item], voc, icons_path)
+            container.pack_start(cpu_avg, False, False, panel["items-padding"])
+
         if item == "dwl-tags":
             if os.path.isfile(common.dwl_data_file):
                 if "dwl-tags" not in panel:
@@ -532,7 +531,7 @@ def main():
     global sig_dwl
     sig_dwl = args.sigdwl
 
-    catchable_sigs = set(signal.Signals) - {signal.SIGKILL, signal.SIGSTOP}
+    catchable_sigs = set(signal.Signals) - {signal.SIGKILL, signal.SIGSTOP, signal.SIGCHLD}
     for sig in catchable_sigs:
         try:
             signal.signal(sig, signal_handler)
@@ -553,9 +552,6 @@ def main():
 
     save_string("-c {} -s {}".format(args.config, args.style), os.path.join(local_dir(), "args"))
 
-    common.app_dirs = get_app_dirs()
-    # common.name2icon_dict = map_odd_desktop_files()
-
     config_file = os.path.join(common.config_dir, args.config)
 
     copy_files(os.path.join(dir_name, "icons_light"), os.path.join(common.config_dir, "icons_light"))
@@ -566,8 +562,10 @@ def main():
     copy_files(os.path.join(dir_name, "config"), common.config_dir, args.restore)
     copy_files(os.path.join(dir_name, "local"), local_dir())
 
-    tree = common.i3.get_tree() if sway else None
-    common.outputs = list_outputs(sway=sway, tree=tree)
+    # tree = common.i3.get_tree() if sway else None
+    common.outputs, common.mon_desc2output_name = list_outputs(sway=sway)
+    print(f"Outputs: {common.outputs}")
+    print(f"Descriptions: {common.mon_desc2output_name}")
 
     panels = load_json(config_file)
 
@@ -590,6 +588,12 @@ def main():
     to_append = []
     for panel in panels:
         check_key(panel, "output", "")
+        check_key(panel, "monitor", "")
+        if panel["monitor"]:
+            try:
+                panel["output"] = common.mon_desc2output_name[panel["monitor"]]
+            except KeyError as err:
+                eprint(f"Monitor description unknown: {err}")
 
         clones = []
         if panel["output"] == "All" and len(common.outputs) >= 1:
@@ -627,6 +631,7 @@ def main():
         # This is to allow width "auto" value. Actually all non-numeric values will be removed.
         if "width" in panel and not isinstance(panel["width"], int):
             panel.pop("width")
+            panel["width-as-percentage"] = False
 
         if panel["output"] in common.outputs or not panel["output"]:
             check_key(panel, "spacing", 6)
@@ -637,6 +642,8 @@ def main():
             check_key(panel, "use-sigrt", False)
             check_key(panel, "start-hidden", False)
 
+            check_key(panel, "width-as-percentage", False)
+
             window = Gtk.Window()
             global panel_windows_hide_show_sigs
             if panel["use-sigrt"]:
@@ -658,6 +665,12 @@ def main():
             if "output" in panel and panel["output"] and "width" not in panel:
                 panel["width"] = common.outputs[panel["output"]]["width"]
 
+            # Width defined as percentage
+            if "width" in panel and panel["width"] > 0 and panel["width-as-percentage"]:
+                if panel["width"] > 100:
+                    panel["width"] = 100
+                panel["width"] = int(common.outputs[panel["output"]]["width"] * panel["width"] / 100)
+
             check_key(panel, "width", 0)
             w = panel["width"]
 
@@ -720,10 +733,6 @@ def main():
 
             inner_box = Gtk.Box(orientation=o, spacing=0)
 
-            # set equal columns width by default if "modules-center" not empty; this may be overridden in config
-            if panel["modules-center"] and panel["homogeneous"]:
-                inner_box.set_homogeneous(True)
-
             hbox.pack_start(inner_box, True, True, 0)
             hbox.set_property("margin-start", panel["padding-horizontal"])
             hbox.set_property("margin-end", panel["padding-horizontal"])
@@ -756,21 +765,29 @@ def main():
                 left_box.pack_start(ms, False, False, 0)
 
             instantiate_content(panel, left_box, panel["modules-left"], icons_path=icons_path)
+            print("left box created")
 
             center_box = Gtk.Box(orientation=o, spacing=panel["spacing"])
             center_box.set_property("name", "center-box")
-            inner_box.pack_start(center_box, True, False, 0)
+
+            if panel["modules-center"] and panel["homogeneous"]:
+                inner_box.set_center_widget(center_box)
+            else:
+                inner_box.pack_start(center_box, True, False, 0)
+
             check_key(panel, "modules-center", [])
             instantiate_content(panel, center_box, panel["modules-center"], icons_path=icons_path)
+            print("center box created")
 
             right_box = Gtk.Box(orientation=o, spacing=panel["spacing"])
             right_box.set_property("name", "right-box")
             # Damn on the guy who invented `pack_start(child, expand, fill, padding)`!
             helper_box = Gtk.Box(orientation=o, spacing=0)
             helper_box.pack_end(right_box, False, False, 0)
-            inner_box.pack_start(helper_box, False, True, 0)
+            inner_box.pack_end(helper_box, False, True, 0)
             check_key(panel, "modules-right", [])
             instantiate_content(panel, right_box, panel["modules-right"], icons_path=icons_path)
+            print("right box created")
 
             if panel["menu-start"] == "right":
                 ms = MenuStart(panel["menu-start-settings"], icons_path=icons_path)
@@ -798,6 +815,7 @@ def main():
             window.add(vbox)
 
             GtkLayerShell.init_for_window(window)
+            GtkLayerShell.set_namespace(window, "nwg-panel")
 
             monitor = None
             try:
@@ -807,11 +825,13 @@ def main():
 
             check_key(panel, "layer", "top")
             o = panel["output"] if "output" in panel else "undefined"
-            print("Panel '{}': output: {}, position: {}, layer: {}, width: {}, height: {}".format(panel["name"], o,
-                                                                                                  panel["position"],
-                                                                                                  panel["layer"],
-                                                                                                  panel["width"],
-                                                                                                  panel["height"]))
+            m = panel["monitor"] if "monitor" in panel else "undefined"
+            print("Panel '{}': output: {}, monitor: {}, position: {}, layer: {}, width: {}, height: {}".format(
+                panel["name"], o, m,
+                panel["position"],
+                panel["layer"],
+                panel["width"],
+                panel["height"]))
 
             if monitor:
                 GtkLayerShell.set_monitor(window, monitor)
@@ -855,7 +875,7 @@ def main():
     if sway:
         common.outputs_num = num_active_outputs(common.i3.get_outputs())
     else:
-        common.outputs = list_outputs(sway=sway, tree=tree, silent=True)
+        common.outputs, common.mon_desc2output_name = list_outputs(sway=sway, silent=True)
         common.outputs_num = len(common.outputs)
 
     if sway:
@@ -873,7 +893,8 @@ def main():
     if his:
         if len(common.h_taskbars_list) > 0 or len(common.h_workspaces_list) > 0:
             print("his: '{}', starting hypr_watcher".format(his))
-            thread = threading.Thread(target=hypr_watcher)
+            # read from Hyprland socket2 on another thread
+            thread = threading.Thread(target=hypr_watcher, daemon=True)
             thread.daemon = True
             thread.start()
 
